#include "sc-include.h"
#include "sc-lib.h"
#include "sc-slist.h"
#include "sc-string.h"

#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

#include "md5.h"

#include <gdbm.h>

#include <sys/types.h>
#include <dirent.h>

#define	SC_PATH_MAX	PATH_MAX

static struct
  {
    GDBM_FILE	db;
    SLIST	dirlist;
    int		errors;
    FILE	*log, *md5;
    char	dbfile[SC_PATH_MAX];
    char	outpath[SC_PATH_MAX], tmpfile[SC_PATH_MAX];
    char	logfile[SC_PATH_MAX], md5file[SC_PATH_MAX];
    size_t	bigbuf_max;
    void	*bigbuf;
    int		dirs, examined, changed, known, added;
    unsigned long long	copy;
  } config;

static void
progress(void)
{
  printf("%d e%d d%d f%d m%d o%d n%d %LdMB\r",
	 (config.dirlist ? config.dirlist->count : 0),
	 config.errors,
	 config.dirs,
	 config.examined,
	 config.changed,
	 config.known,
	 config.added,
	 config.copy>>20);
  fflush(stdout);
}

static void
vlog(const char *prefix, const char *s, va_list list, const char *suffix, ...)
{
  static time_t	last;
  time_t	tim;
  struct tm	tm;
  va_list	tmp;

  time(&tim);
  if (last!=tim)
    progress();
  last	= tim;
  gmtime_r(&tim, &tm);

  if (!config.log)
    return;
  fprintf(config.log,
	  "%4d-%02d-%02d %02d:%02d:%02d %s",
	  1900+tm.tm_year, tm.tm_mon, tm.tm_mday,
	  tm.tm_hour, tm.tm_min, tm.tm_sec, prefix);
  vfprintf(config.log, s, list);
  fprintf(config.log, "%s", prefix);
  va_start(tmp, suffix);
  vfprintf(config.log, suffix, tmp);
  va_end(tmp);
  fputc('\n', config.log);
  fflush(config.log);
}

static void
log(const char *s, ...)
{
  va_list	list;

  va_start(list, s);
  vlog("", s, list, "");
  va_end(list);
}

static void
verror(const char *prefix, const char *s, va_list list)
{
  const char	*err;

  err	= strerror(errno);

  config.errors++;

  vlog(prefix, s, list, ": %s", err);

  fputs(prefix, stderr);
  vfprintf(stderr, s, list);
  fprintf(stderr, ": %s\n", err);
}

static void
error(const char *s, ...)
{
  va_list	list;

  va_start(list, s);
  verror("error: ", s, list);
  va_end(list);
}

static void
fatal(const char *s, ...)
{
  va_list	list;

  va_start(list, s);
  verror("fatal: ", s, list);
  va_end(list);
  exit(-1);
  abort();
  for(;;);
}

static void
assemble(char *s, size_t max, const char *fmt, ...)
{
  int		n;
  va_list	list;

  va_start(list, fmt);
  n	= vsnprintf(s, max, fmt, list);
  va_end(list);
  if (n>=max)
    fatal("argument too long");
}

struct sc_dir_stat
  {
    const char	*name;	/* Full path	*/
    struct stat64	st;
  };

#define	SC_PATH_MAX	PATH_MAX

/* This gobbles CWD!
 */
static GLIST
sc_scandir(const char *name)
{
  SLIST		slist;
  GLIST		glist;
  DIR		*dir;
  struct dirent	*dp;
  char		buf[SC_PATH_MAX+2];
  const char	*s;

  xDP(("sc_scandir(%s)", name));
  /* Read directory contents
   */
  if ((dir=opendir(name))==NULL)
    {
      error("cannot open dir: %s", name);
      return 0;
    }
  slist	= slist_new();
  while ((dp=readdir(dir))!=NULL)
    if (!(dp->d_name[0]=='.' &&
	  (dp->d_name[1]==0 ||
	   (dp->d_name[1]=='.' && dp->d_name[2]==0))))
	{
	  xDP(("sc_scandir() readdir %s", dp->d_name));
	  slist_add(slist, dp->d_name);
	}
  if (closedir(dir))
    error("error reading dir: %s", name);

  /* Now do the stats
   */
  glist	= glist_new(sizeof (struct sc_dir_stat));
  for (; (s=slist_get(slist))!=NULL; slist_free(s))
    {
      struct stat64		st;
      struct sc_dir_stat	*ent;

      xDP(("sc_scandir() stat %s", s));
      strxcpy(buf, name, sizeof buf);
      if (!*buf || buf[strlen(buf)-1]!='/')
	strxcat(buf, "/", sizeof buf);
      strxcat(buf, s, sizeof buf);
      if (strlen(buf)>=SC_PATH_MAX)
	{
	  error("filename too deep: %s/%s", name, s);
	  continue;
	}
      if (lstat64(buf, &st))
	{
	  error("cannot stat: %s", buf);
	  continue;
	}
      ent	= glist_add(glist)->data;
      ent->name	= str_alloc(buf);
      ent->st	= st;
    }
  xDP(("sc_scandir() ok"));
  return glist;
}

struct data
  {
    unsigned long long	mtime;
    int			count;
    char		md5[34];
  };

static int
dbfetch(const char *name, struct data *data)
{
  char	tmp[32+32+32];
  datum	d;

  config.examined++;
  d.dptr	= (char *)name;
  d.dsize	= strlen(name);
  d	= gdbm_fetch(config.db, d);
  if (d.dptr==0)
    return 0;
  xDP(("db(%s) %d '%.*s'", name, d.dsize, d.dsize, d.dptr));
  if (d.dsize<sizeof tmp)
    {
      unsigned long long	tim, cnt;
      int			p1, p2;

      memcpy(tmp, d.dptr, d.dsize);
      free(d.dptr);
      tmp[d.dsize]	= 0;

      p1=0;
      p2=0;
      sscanf(tmp, "%Lu %Lu %n%*s%n", &tim, &cnt, &p1, &p2);
      if (3<p1 && p1+32==p2)
	{
	  data->mtime	= tim;
	  data->count	= (int)cnt;
	  memcpy(data->md5, tmp+p1, 32);
	  data->md5[32]	= 0;
	  return 1;
	}
    }
  error("database corrupt for %s", name);
  return 0;
}

static int
dbstore(const char *name, struct data *data)
{
  char	tmp[32+32+32];
  datum	k,v;

  assemble(tmp, sizeof tmp,
	   "%Lu %Lu %s",
	   (unsigned long long)data->mtime,
	   (unsigned long long)data->count,
	   data->md5);
  k.dptr	= (char *)name;
  k.dsize	= strlen(name);
  v.dptr	= tmp;
  v.dsize	= strlen(tmp);
  if (!gdbm_store(config.db, k, v, GDBM_REPLACE))
    return 1;
  error("database not writeable for %s", name);
  return 0;
}

static int
readit(int fd, char *buf, size_t len)
{
  int	got;

  while ((got=read(fd, buf, len))<0)
    if (errno!=EINTR && errno!=EAGAIN)
      break;
  return got;
}

static int
writeall(int fd, const char *buf, size_t len)
{
  int	pos, put;

  for (pos=0; pos<len; )
    {
      put	= write(fd, buf+pos, len-pos);
      if (put<=0)
	{
	  if (!put)
	    break;
	  if (errno!=EINTR && errno!=EAGAIN)
	    break;
	  continue;
	}
      if (put>len-pos)
	fatal("syscall defect, wrote more than allowed");
      pos+=put;
    }
  return pos;
}

static int
readcmp(int fd, const char *buf, size_t len)
{
  int	pos, got, max;
  char	blk[BUFSIZ*16];

  for (pos=0; pos<len; pos+=got)
    {
      max	= len-pos;
      if (max>sizeof blk)
	max	= sizeof blk;
      got	= readit(fd, blk, max);
      if (got<=0)
	{
	  if (got)
	    return -1;
	  break;
	}

      if (got>max)
	fatal("syscall defect, read more than allowed");

      if (memcmp(buf+pos, blk, got))
	return 1;
    }
  return 0;
}

/* Do a copy from source to tmpfile
 * with an md5 sum in between.
 */
static int
md5copy(const char *name, char *md5)
{
  MD5_CTX	ctx;
  char		digest[16];
  int		got, i;
  int		in, out;
  int		errstate;

  /* Cache the output as it's likely we must compare it
   */
  if ((out=open(config.tmpfile, O_WRONLY|O_CREAT|O_TRUNC|O_LARGEFILE, 0775))<0)
    {
      error("cannot create temporary %s", config.tmpfile);
      return 0;
    }
  /* Don't cache the input as it's not likely read again
   */
  if ((in=open(name, O_RDONLY|O_NOFOLLOW|O_DIRECT|O_LARGEFILE))<0)
    {
      close(out);
      error("cannot open %s", name);
      return 0;
    }

  MD5_Init(&ctx);

  errstate	= 0;
  while ((got=readit(in, config.bigbuf, config.bigbuf_max))>0)
    {
      MD5_Update(&ctx, config.bigbuf, got);
      if (writeall(out, config.bigbuf, got)!=got)
	{
	  if (errno==ENOSPC)
	    {
	      unlink(config.tmpfile);
	      errno	= ENOSPC;
	      fatal("disk full at tmpfile: %s", config.tmpfile);
	    }

	  /* Hm .. what kind of nonpermanent reoccurring error?
	   * NFS softmount dead?
	   */
	  error("cannot write to tmpfile: %s", config.tmpfile);
	  errstate	= 1;
	  break;
	}
      config.copy	+= got;
    }
  if (got<0)
    {
      error("read error on input: %s", name);
      errstate	= 1;
    }
  if (close(in))
    {
      error("error closing input: %s", name);
      errstate	= 1;
    }
  if (close(out))
    {
      error("error closing tmpfile: %s", config.tmpfile);
      errstate	= 1;
    }
  if (errstate)
    {
      unlink(config.tmpfile);
      return 0;
    }

  MD5_Final(digest, &ctx);

  for (i=0; i<16; i++)
    {
      *md5++	= "0123456789abcdef"[(digest[i]>>4)&0xf];
      *md5++	= "0123456789abcdef"[(digest[i]>>0)&0xf];
    }
  *md5	= 0;
  return 1;
}

/* Write the md5sum compatible log entry
 * Complex, as md5sum may use escaped syntax.
 */
static void
md5log(const char *name, const char *md5)
{
  int	i;

  /* Escaped syntax on files containing \n or \\
   */
  for (i=0; name[i]; i++)
    if (name[i]=='\\' || name[i]=='\n')
      {
	fprintf(config.md5, "\\%s *", md5);
	for (i=0; name[i]; i++)
	  if (name[i]=='\n')
	    {
	      fputc('\\', config.md5);
	      fputc('n', config.md5);
	      continue;
	    }
	  else
	    {
	      if (name[i]=='\\')
		fputc('\\', config.md5);
	      fputc(name[i], config.md5);
	    }
	fputc('\n', config.md5);
      }
  /* Ordinary output
   */
  fprintf(config.md5, "%s *%s\n", md5, name);
}

/* Compare open FD to backup tmpfile
 *
 * Returns:
 * <0	on error
 * 0	if identical
 * >0	if differs
 */
static int
compare_file(int fd, const char *name)
{
  int	cmp;
  int	res;
  int	got;

  xDP(("compare_file(%d,%s)", fd,name));

  if ((cmp=open(config.tmpfile, O_RDONLY|O_LARGEFILE, 0775))<0)
    fatal("cannot read tmpfile: %s", config.tmpfile);

  res	= 0;
  while ((got=readit(fd, config.bigbuf, config.bigbuf_max))>0)
    if ((res=readcmp(cmp, config.bigbuf, got))!=0)
      {
	if (res<0)
	  fatal("read error: %s", config.tmpfile);
	break;
      }
  if (got<0)
    return -1;
  if (!res)
    {
      char	tmp[BUFSIZ];

      /* Check for EOF on both files
       */
      if ((res=readit(cmp, tmp, sizeof tmp))<0)
	fatal("read error: %s", config.tmpfile);
    }
  if (close(cmp))
    fatal("close error: %s", config.tmpfile);
  return res;
}

/* Returns:
 * <0	hardlink failed or retryable error
 * ==0	mismatch (funny, I would rather suspect a corrupt file system)
 * >0	match or successful hardlink done
 */
static int
compare_or_link(const char *md5, int count)
{
  char	out[SC_PATH_MAX];
  char	tmp[SC_PATH_MAX];
  int	in;

  /* XXX
   * Check for compressed here in future
   */

  xDP(("compare_or_link(%s,%d)", md5,count));

  assemble(out, sizeof out,
	   "%s/%c%c/%c%c/%s.%d.md5",
	   config.outpath,
	   md5[0], md5[1],
	   md5[2], md5[3],
	   md5, count);

  if ((in=open(out, O_RDONLY|O_NOFOLLOW|O_DIRECT|O_LARGEFILE))>=0)
    {
      int	diff;
      
      diff	= compare_file(in, out);

      /* In case of error we really don't want to continue
       * as something must be badly broken on the target.
       */
      if (diff<0 || close(in))
	fatal("read error in backup file: %s", out);

      if (diff)
	return 0;

      config.known++;
      return 1;
    }

  if (errno!=ENOENT)
    {
      /* Quite not funny
       * We hit some unexpected error.
       * Be conservative and don't try to disturb any more
       */
      fatal("unexpected error accessing file: %s", out);
    }

  if (!link(config.tmpfile, out))
    {
      config.added++;
      return 1;
    }

  if (errno!=ENOENT)
    return -1;

  /* Try to create missing directories
   */
  assemble(tmp, sizeof tmp,
	   "%s/%c%c",
	   config.outpath,
	   md5[0], md5[1]);
  if (!mkdir(tmp, 0775))
    log("created directory %s", tmp);

  assemble(tmp, sizeof tmp,
	   "%s/%c%c/%c%c",
	   config.outpath,
	   md5[0], md5[1],
	   md5[2], md5[3]);
  if (!mkdir(tmp, 0775))
    log("created directory %s", tmp);

  if (!link(config.tmpfile, out))
    {
      config.added++;
      return 1;
    }

  return -1;
}

/* Hunt for existing MD5 files.
 * Or move it yourself.
 */
static int
md5hunt(const char *md5)
{
  int	count;

  xDP(("md5hunt(%s)", md5));
  for (count=0;; count++)
    {
      int	ret, reloop;

      /* According to the NFS documentation
       * the NFS server might go out of business
       * right before telling us that a hardlink was done.
       * So we have to reloop a bit here.
       */
      for (reloop=0; (ret=compare_or_link(md5, count))<0; reloop++)
	if (reloop>4)
	  fatal("hardlink failed: %s %d", md5, count);

      /* We had a match,
       */
      if (ret>0)
	break;
    }
  /* In either case,
   * we do not need the tmpfile any more
   */
  unlink(config.tmpfile);
  return count;
}

/* Do the backup
 */
static void
backup(const char *name, unsigned long long mtime)
{
  struct data	d;
  char		md5[34];

  xDP(("backup(%s,%Lu)", name, mtime));
  if (dbfetch(name, &d) && d.mtime==mtime)
    {
      xDP(("backup() not newer"));
      return;
    }
  config.changed++;
#if 0
  log("backup %s", name);
#endif
  if (md5copy(name, md5))
    {
      int	cnt;

      xDP(("backup() copied %s", md5));
      md5log(name, md5);
      cnt	= md5hunt(md5);
      if (cnt>=0)
	{
	  d.mtime	= mtime;
	  d.count	= cnt;
	  strcpy(d.md5, md5);
	  if (dbstore(name, &d))
	    {
	      log("md5 %s %d %s", md5, cnt, name);
	      return;
	    }
	}
    }
  log("failed backup %s", name);
}


static void
md5backup(int argc, char **argv)
{
  const char	*s;

  xDP(("md5backup(%d,%p)", argc, argv));
  for (config.dirlist=slist_init(argc, argv);
       (s=slist_get(config.dirlist))!=0; slist_free(s))
    {
      GLIST			glist;
      GLIST_ENT			ent;

      config.dirs++;
      progress();

      glist	= sc_scandir(s);
      xDP(("md5backup() glist=%p", glist));
      while ((ent=glist_get(glist))!=0)
	{
	  struct sc_dir_stat	*ptr=ent->data;

	  xDP(("md5backup() name=%s", ptr->name));
	  if (S_ISDIR(ptr->st.st_mode))
	    slist_add(config.dirlist, ptr->name);
	  else if (S_ISREG(ptr->st.st_mode))
	    backup(ptr->name, ptr->st.st_mtime);
	  glist_free(ent);
	}
      xDP(("md5backup() next"));
    }
}

static int
existsdir(const char *dir)
{
  struct stat	s;

  if (stat(dir, &s))
    return 0;
  return S_ISDIR(s.st_mode);
}

static void
init(const char *path, const char *node)
{
  assemble(config.logfile, sizeof config.logfile,
	   "%slog/%s", path, node);
  if ((config.log=fopen(config.logfile, "a"))==0)
    fatal("cannot append to logfile %s", config.logfile);

  log("start %s %s", path, node);

  assemble(config.outpath, sizeof config.outpath,
	   "%sout", path);
  if (!existsdir(config.outpath))
    fatal("output directory does not exist: %s", config.outpath);

  assemble(config.tmpfile, sizeof config.tmpfile,
	   "%stmp/%s.%d", path, node, getpid());

  assemble(config.md5file, sizeof config.md5file,
	   "%smd5/%s", path, node);
  if ((config.md5=fopen(config.md5file, "a"))==0)
    fatal("cannot append to md5sum %s", config.md5file);

  assemble(config.dbfile, sizeof config.dbfile,
	   "%sdbm/%s", path, node);
  if ((config.db=gdbm_open(config.dbfile,
			   BUFSIZ, GDBM_WRCREAT|GDBM_SYNC, 0775, NULL))==0)
    fatal("cannot open database: %s", config.dbfile);

  config.bigbuf_max	= 1<<20;
  config.bigbuf		= alloc(config.bigbuf_max);
}

static void
cleanup(void)
{
  FILE	*tmp;

  gdbm_close(config.db);

  if (fclose(config.md5))
    error("close md5sum %s", config.md5file);

  unlink(config.tmpfile);

  log("ready, err=%d dir=%d file=%d mod=%d old=%d new=%d",
      config.errors,
      config.dirs,
      config.examined,
      config.changed,
      config.known,
      config.added);

  printf("err=%d dir=%d file=%d mod=%d old=%d new=%d\n",
	 config.errors,
	 config.dirs,
	 config.examined,
	 config.changed,
	 config.known,
	 config.added);

  tmp	= config.log;
  config.log	= 0;
  if (fclose(tmp))
    error("close logfile %s", config.logfile);
}

int
main(int argc, char **argv)
{
  if (argc<3)
    {
      printf("Usage: %s /target/path nodename dir...\n"
	     , argv[0]);
      return 1;
    }
  init(argv[1], argv[2]);
  md5backup(argc-3, argv+3);
  cleanup();
  return 0;
}
